Index: apps/app_queue.c
===================================================================
--- apps/app_queue.c	(revision 46429)
+++ apps/app_queue.c	(working copy)
@@ -459,14 +459,40 @@
 }
 
 struct statechange {
+	AST_LIST_ENTRY(statechange) list;
 	int state;
 	char dev[0];
 };
 
-static void *changethread(void *data)
+static AST_LIST_HEAD_STATIC(device_state_changes, statechange);
+static pthread_t change_thread = AST_PTHREADT_NULL;
+static ast_cond_t device_change_pending;
+static void *do_device_change(const struct statechange *sc);
+
+static void *changethread(void *unused)
 {
+	struct statechange *sc = NULL;
+
+	AST_LIST_LOCK(&device_state_changes);
+	for (;;) {
+		sc = AST_LIST_REMOVE_HEAD(&device_state_changes, list);
+		if (sc) {
+			AST_LIST_UNLOCK(&device_state_changes);
+			do_device_change(sc);
+			free(sc);
+			AST_LIST_LOCK(&device_state_changes);
+		} else {
+			ast_cond_wait(&device_change_pending, &device_state_changes.lock);
+		}
+	}
+
+	return NULL;
+}
+
+
+static void *do_device_change(const struct statechange *sc)
+{
 	struct call_queue *q;
-	struct statechange *sc = data;
 	struct member *cur;
 	struct member_interface *curint;
 	char *loc;
@@ -477,7 +503,6 @@
 	if (loc) {
 		*loc++ = '\0';
 	} else {
-		free(sc);
 		return NULL;
 	}
 
@@ -498,7 +523,6 @@
 	if (!curint) {
 		if (option_debug)
 			ast_log(LOG_DEBUG, "Device '%s/%s' changed to state '%d' (%s) but we don't care because they're not a member of any queue.\n", technology, loc, sc->state, devstate2str(sc->state));
-		free(sc);
 		return NULL;
         }
 
@@ -548,21 +572,18 @@
 	/* Avoid potential for deadlocks by spawning a new thread to handle
 	   the event */
 	struct statechange *sc;
-	pthread_t t;
-	pthread_attr_t attr;
 
-	if (!(sc = malloc(sizeof(*sc) + strlen(dev) + 1)))
+	if (!(sc = calloc(1, sizeof(*sc) + strlen(dev) + 1)))
 		return 0;
 
 	sc->state = state;
 	strcpy(sc->dev, dev);
-	pthread_attr_init(&attr);
-	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
-	if (ast_pthread_create(&t, &attr, changethread, sc)) {
-		ast_log(LOG_WARNING, "Failed to create update thread!\n");
-		free(sc);
-	}
 
+	AST_LIST_LOCK(&device_state_changes);
+	AST_LIST_INSERT_TAIL(&device_state_changes, sc, list);
+	ast_cond_signal(&device_change_pending);
+	AST_LIST_UNLOCK(&device_state_changes);
+
 	return 0;
 }
 
@@ -4002,7 +4023,12 @@
 int load_module(void)
 {
 	int res;
-	
+
+	if (change_thread == AST_PTHREADT_NULL) {
+		ast_cond_init(&device_change_pending, NULL);
+		ast_pthread_create(&change_thread, NULL, changethread, NULL);
+	}
+
 	res = ast_register_application(app, queue_exec, synopsis, descrip);
 	res |= ast_cli_register(&cli_show_queue);
 	res |= ast_cli_register(&cli_show_queues);
